/*!
  * @license MIT (https://github.com/geotrev/rotom/blob/master/LICENSE)
  * Rotom v0.12.0 (https://github.com/geotrev/rotom#readme)
  * Copyright 2022 George Treviranus <geowtrev@gmail.com>
  */
import { patch, create, render } from 'omdomdom';

const createUUID = () => {
  const base = Number.MAX_SAFE_INTEGER;
  return Math.floor(Math.random() * base).toString(36) + Math.abs(Date.now()).toString(36);
};

const getTypeTag = value => Object.prototype.toString.call(value).slice(8, -1).toLowerCase();
const isPlainObject = value => getTypeTag(value) === "object";
const isEmptyObject = value => {
  if (!isPlainObject(value)) return false;

  for (let key in value) {
    if (Object.prototype.hasOwnProperty.call(value, key)) {
      return false;
    }
  }

  return true;
};
const isFunction = value => getTypeTag(value) === "function";
const isString = value => getTypeTag(value) === "string";
const isUndefined = value => getTypeTag(value) === "undefined";

const camelToKebab = value => value && value.match(/[A-Z]{2,}(?=[A-Z][a-z]+[0-9]*|\b)|[A-Z]?[a-z]+[0-9]*|[A-Z]|[0-9]+/g).map(x => x.toLowerCase()).join("-");
const kebabToCamel = value => value && value.split("-").map((word, i) => i > 0 ? word[0].toUpperCase() + word.slice(1).toLowerCase() : word.toLowerCase()).join("");

const sanitizeString = value => {
  const dump = document.createElement("div");
  dump.textContent = value;
  return "" + dump.innerHTML;
};

function log(msg) {
  let type = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "warn";
  console[type]("[RotomElement]: ".concat(msg));
}

const Attributes = {
  dir: "dir"
};
const AttributeValues = {
  ltr: "ltr"
};
const External = {
  rotomIdProperty: "rotomId",
  requestRender: "requestRender",
  render: "render",
  staticProperties: "properties",
  staticStyles: "styles",
  rotomIdAttribute: "rotom-id",
  onConnect: "onConnect",
  onMount: "onMount",
  onUpdate: "onUpdate",
  onPropertyChange: "onPropertyChange",
  onAttributeChange: "onAttributeChange",
  onUnmount: "onUnmount"
};
const Internal = {
  rotomId: Symbol("#rotomId"),
  vnode: Symbol("#vnode"),
  isFirstRender: Symbol("#isFirstRender"),
  reflectMap: Symbol("#reflectMap"),
  renderer: Symbol("#renderer"),
  initialize: Symbol("#initialize"),
  schedule: Symbol("#schedule"),
  runLifecycle: Symbol("#runLifecycle"),
  upgrade: Symbol("#upgrade"),
  upgradeProperties: Symbol("#upgradeProperties"),
  renderStyles: Symbol("#renderStyles"),
  renderDOM: Symbol("#renderDOM"),
  patch: Symbol("#patch"),
  destroy: Symbol("#destroy")
};

function setRenderer(renderer) {
  return {
    patch(element) {
      if (!window || !window.document) return;

      if (!isFunction(element[External.render])) {
        throw new Error("[RotomElement]: You must include a render method in element: '".concat(element.constructor.name, "'"));
      }

      renderer.patch(element);

      if (element[Internal.isFirstRender]) {
        element[Internal.isFirstRender] = false;
        element[Internal.runLifecycle](External.onMount);
      } else {
        element[Internal.runLifecycle](External.onUpdate);
      }
    },

    destroy(element) {
      if (!window || !window.document) return;
      element[Internal.isFirstRender] = true;
      element[Internal.runLifecycle](External.onUnmount);
      renderer.destroy(element);
      element[Internal.vnode] = null;
      const children = Array.apply(null, element.shadowRoot.childNodes);

      if (children.length) {
        Array.prototype.forEach.call(children, child => element.shadowRoot.removeChild(child));
      }
    }

  };
}

function createScheduler() {
  const requestAnimationFrame = window.requestAnimationFrame;
  const setTimeout = window.setTimeout;
  const FRAME_DURATION = 1000 / 60;
  let scheduled = null;

  function schedule(fn) {
    if (scheduled) return;

    if (isFunction(requestAnimationFrame)) {
      scheduled = requestAnimationFrame(() => {
        scheduled = null;
        fn();
      });
    } else if (isFunction(setTimeout)) {
      scheduled = setTimeout(() => {
        scheduled = null;
        fn();
      }, FRAME_DURATION);
    }
  }

  return fn => schedule(fn);
}

function validateType(name, value, type) {
  if (isUndefined(type)) return;
  const evaluatedType = getTypeTag(value);

  if (evaluatedType !== type) {
    log("Property '".concat(name, "' is type '").concat(evaluatedType, "', expected '").concat(type, "'."));
  }
}

function validateRequired(name, value, type, required) {
  if (!required) return;

  if (isUndefined(value)) {
    if (isUndefined(type)) {
      log("Property '".concat(name, "' is required."), "error");
    } else {
      log("Property '".concat(name, "' of type '").concat(type, "' is required."), "error");
    }
  }
}

function setDefaultvalue(Cls, propName, _ref, privateName) {
  let {
    default: defaultValue,
    type: propType,
    reflected = false,
    safe = false,
    required = false
  } = _ref;
  const attrName = camelToKebab(propName);
  const attrValue = Cls.getAttribute(attrName);
  const initialValue = reflected && attrValue ? attrValue : !isUndefined(Cls[propName]) ? Cls[propName] : isFunction(defaultValue) ? defaultValue(Cls) : defaultValue;

  {
    validateType(propName, initialValue, propType);
    validateRequired(propName, initialValue, propType, required);
  }

  if (!isUndefined(initialValue)) {
    const privateValue = safe && isString(initialValue) ? sanitizeString(initialValue) : initialValue;
    Cls[privateName] = privateValue;
  }

  if (reflected && !attrValue) {
    const initialAttrValue = initialValue ? String(initialValue) : "";
    Cls.setAttribute(attrName, initialAttrValue);
  }
}

function upgradeProperty(Cls, propName) {
  let configuration = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

  if (Object.getOwnPropertyDescriptor(Object.getPrototypeOf(Cls), propName)) {
    return;
  }

  const privateName = Symbol(propName);
  const {
    type,
    reflected = false,
    safe = false,
    required = false
  } = configuration;

  if (reflected) {
    Cls[Internal.reflectMap][propName] = privateName;
  }

  setDefaultvalue(Cls, propName, configuration, privateName);
  Object.defineProperty(Cls, propName, {
    configurable: true,
    enumerable: true,

    get() {
      return Cls[privateName];
    },

    set(value) {
      if (value === Cls[privateName]) return;

      {
        validateType(propName, value, type);
        validateRequired(propName, value, type, required);
      }

      if (reflected && !Cls[Internal.reflectMap][propName]) {
        Cls[Internal.reflectMap][propName] = privateName;
      }

      const oldValue = Cls[privateName];

      if (!isUndefined(value)) {
        Cls[privateName] = safe && type === "string" && isString(value) && value !== "" ? sanitizeString(value) : value;
        Cls[Internal.runLifecycle](External.onPropertyChange, propName, oldValue, value);

        if (reflected) {
          const attribute = camelToKebab(propName);
          const attrValue = String(value);

          if (Cls.getAttribute(attribute) !== attrValue) {
            Cls.setAttribute(attribute, attrValue);
          }
        }
      } else {
        delete Cls[privateName];
        Cls[Internal.runLifecycle](External.onPropertyChange, propName, oldValue, value);

        if (reflected) {
          delete Cls[Internal.reflectMap][propName];
          const attribute = camelToKebab(propName);
          Cls.removeAttribute(attribute);
        }
      }

      Cls[External.requestRender]();
    }

  });
}

const SHADOW_ROOT_MODE = "open";
function rotomFactory(renderer) {
  return class RotomElement extends HTMLElement {
    constructor() {
      super();
      this[Internal.schedule] = createScheduler();
      this[Internal.renderer] = setRenderer(renderer);
      this.attachShadow({
        mode: SHADOW_ROOT_MODE
      });
      this[Internal.patch] = this[Internal.patch].bind(this);
      this[Internal.isFirstRender] = true;
      this[Internal.vnode] = null;
      this[Internal.rotomId] = createUUID();
      this[Internal.reflectMap] = {};
    }

    static get observedAttributes() {
      const properties = this[External.staticProperties];
      if (isEmptyObject(properties)) return [];
      let attributes = [];

      for (let propName in properties) {
        if (!properties[propName].reflected) continue;
        attributes.push(camelToKebab(propName));
      }

      return attributes;
    }

    adoptedCallback() {}

    attributeChangedCallback(name, oldValue, newValue) {
      if (oldValue !== newValue) {
        this[Internal.runLifecycle](External.onAttributeChange, name, oldValue, newValue);
        const propName = kebabToCamel(name);
        const privateName = this[Internal.reflectMap][propName];

        if (!isUndefined(privateName) && this[propName] !== newValue) {
          this[propName] = newValue;
        }
      }
    }

    connectedCallback() {
      if (!this.isConnected) return;
      this[Internal.upgrade]();
      this[Internal.runLifecycle](External.onConnect);
      this[Internal.renderStyles]();
      this[External.requestRender]();
    }

    disconnectedCallback() {
      this[Internal.destroy]();
    }

    get [External.rotomIdProperty]() {
      return this[Internal.rotomId];
    }

    [External.requestRender]() {
      this[Internal.schedule](this[Internal.patch]);
    }

    [Internal.runLifecycle](methodName) {
      if (isFunction(this[methodName])) {
        for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
          args[_key - 1] = arguments[_key];
        }

        this[methodName](...args);
      }
    }

    [Internal.upgrade]() {
      this.setAttribute(External.rotomIdAttribute, this[External.rotomIdProperty]);
      this.setAttribute(Attributes.dir, String(document.dir || AttributeValues.ltr));
      this[Internal.upgradeProperties]();
    }

    [Internal.upgradeProperties]() {
      const properties = this.constructor[External.staticProperties];
      if (isEmptyObject(properties)) return;

      for (let propName in properties) {
        upgradeProperty(this, propName, properties[propName]);
      }
    }

    [Internal.destroy]() {
      this[Internal.renderer].destroy(this);
    }

    [Internal.patch]() {
      this[Internal.renderer].patch(this);
    }

    [Internal.renderStyles]() {
      const styles = this.constructor[External.staticStyles];
      if (!isString(styles)) return;
      const styleTag = document.createElement("style");
      styleTag.type = "text/css";
      styleTag.textContent = styles;
      this.shadowRoot.appendChild(styleTag);
    }

  };
}

const emptyVNode = {
  type: "comment",
  attributes: {},
  children: null,
  content: "",
  node: document.createComment("")
};

function getRenderState(element) {
  const domString = element[External.render]();

  if (!isString(domString)) {
    throw new Error("[RotomElement]: You attempted to render a non-string template in element: '".concat(element.constructor.name, "'."));
  }

  return domString;
}

function getInitialRenderState(element) {
  element[Internal.vnode] = create(getRenderState(element));
  render(element[Internal.vnode], element.shadowRoot);
}

function getNextRenderState(element) {
  let nextVnode = create(getRenderState(element));
  patch(nextVnode, element[Internal.vnode]);
  nextVnode = null;
}

const templateRenderer = {
  patch(element) {
    if (element[Internal.isFirstRender]) {
      getInitialRenderState(element);
    } else {
      getNextRenderState(element);
    }
  },

  destroy(element) {
    patch(emptyVNode, element[Internal.vnode]);
  }

};

function createDirectionObserver() {
  if (window.__ROTOM_ELEMENT__DIR_OBSERVER__) return;
  window.__ROTOM_ELEMENT__DIR_OBSERVER__ = true;

  const updateDirection = function () {
    let context = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : document;
    const nodes = Array.apply(null, context.querySelectorAll("[rotom-id]"));
    if (!nodes.length) return;
    nodes.forEach(node => {
      node.setAttribute("dir", String(document.dir || "ltr"));

      if (node.shadowRoot) {
        updateDirection(node.shadowRoot);
      }
    });
  };

  const mutationObserver = new MutationObserver(mutations => {
    mutations.forEach(mutation => {
      if (mutation.attributeName === "dir") {
        updateDirection();
      }
    });
  });
  mutationObserver.observe(document.documentElement, {
    attributes: true
  });
}

const register = (tag, RotomInstance) => {
  if (!customElements.get(tag)) {
    customElements.define(tag, RotomInstance);
  }
};

createDirectionObserver();
const RotomElement = rotomFactory(templateRenderer);

export { RotomElement, register, validateType };
//# sourceMappingURL=rotom.template.esm.js.map
